# Unit test _predict_interval_conformal ForecasterDirect
# ==============================================================================
import numpy as np
import pandas as pd
from sklearn.compose import ColumnTransformer
from sklearn.preprocessing import StandardScaler
from sklearn.preprocessing import OneHotEncoder
from sklearn.linear_model import LinearRegression
from skforecast.direct import ForecasterDirect

# Fixtures
from .fixtures_forecaster_direct import y


def test_predict_interval_conformal_output_when_forecaster_is_LinearRegression_steps_is_1_in_sample_residuals_is_True():
    """
    Test output when estimator is LinearRegression and one step ahead is predicted
    using in sample residuals.
    """
    forecaster = ForecasterDirect(
        estimator=LinearRegression(), lags=2, steps=3
    )
    forecaster.fit(y=pd.Series(np.arange(10)), store_in_sample_residuals=True)
    forecaster.in_sample_residuals_ = np.array([10, 20])
    results = forecaster._predict_interval_conformal(
        steps=1, nominal_coverage=0.95, use_in_sample_residuals=True, use_binned_residuals=False
    )

    expected = pd.DataFrame(
                    data    = np.array([[10. , -9.5, 29.5]]),
                    columns = ['pred', 'lower_bound', 'upper_bound'],
                    index   = pd.RangeIndex(start=10, stop=11, step=1)
                )

    pd.testing.assert_frame_equal(results, expected)

    
def test_predict_interval_conformal_output_when_forecaster_is_LinearRegression_steps_is_2_in_sample_residuals_is_True():
    """
    Test output when estimator is LinearRegression and two step ahead is predicted
    using in sample residuals.
    """
    forecaster = ForecasterDirect(
        estimator=LinearRegression(), lags=2, steps=3
    )
    forecaster.fit(y=pd.Series(np.arange(10)), store_in_sample_residuals=True)
    forecaster.in_sample_residuals_ = np.array([10, 20])
    results = forecaster._predict_interval_conformal(
        steps=2, nominal_coverage=0.95, use_in_sample_residuals=True, use_binned_residuals=False
    )

    expected = pd.DataFrame(
                   data    = np.array([[10. , -9.5, 29.5],
                                       [11. , -8.5, 30.5]]),
                   columns = ['pred', 'lower_bound', 'upper_bound'],
                   index   = pd.RangeIndex(start=10, stop=12, step=1)
               )

    pd.testing.assert_frame_equal(results, expected)
    
    
def test_predict_interval_conformal_output_when_forecaster_is_LinearRegression_steps_is_1_in_sample_residuals_is_False():
    """
    Test output when estimator is LinearRegression and one step ahead is predicted
    using out sample residuals.
    """
    forecaster = ForecasterDirect(
        estimator=LinearRegression(), lags=2, steps=3
    )
    forecaster.fit(y=pd.Series(np.arange(10)), store_in_sample_residuals=True)
    forecaster.out_sample_residuals_ = np.array([10, 20])
    results = forecaster._predict_interval_conformal(
        steps=1, nominal_coverage=0.95, use_in_sample_residuals=False, use_binned_residuals=False
    )

    expected = pd.DataFrame(
                   data    = np.array([[10., -9.5, 29.5]]),
                   columns = ['pred', 'lower_bound', 'upper_bound'],
                   index   = pd.RangeIndex(start=10, stop=11, step=1)
               )

    pd.testing.assert_frame_equal(results, expected)
    
    
def test_predict_interval_conformal_output_when_forecaster_is_LinearRegression_steps_is_2_in_sample_residuals_is_False():
    """
    Test output when estimator is LinearRegression and two step ahead is predicted
    using out sample residuals.
    """
    forecaster = ForecasterDirect(
        estimator=LinearRegression(), lags=2, steps=3
    )
    forecaster.fit(y=pd.Series(np.arange(10)), store_in_sample_residuals=True)
    forecaster.out_sample_residuals_ = np.array([10, 20])
    results = forecaster._predict_interval_conformal(
        steps=2, nominal_coverage=0.95, use_in_sample_residuals=False, use_binned_residuals=False
    )

    expected = pd.DataFrame(
                   data    = np.array([[10. , -9.5, 29.5],
                                       [11. , -8.5, 30.5]]),
                   columns = ['pred', 'lower_bound', 'upper_bound'],
                   index   = pd.RangeIndex(start=10, stop=12, step=1)
               )

    pd.testing.assert_frame_equal(results, expected)


def test_predict_interval_conformal_output_when_with_transform_y():
    """
    Test predict output when using LinearRegression as estimator and StandardScaler.
    """
    y = pd.Series(
            np.array([-0.59,  0.02, -0.9 ,  1.09, -3.61,  0.72, -0.11, -0.4 ,  0.49,
                       0.67,  0.54, -0.17,  0.54,  1.49, -2.26, -0.41, -0.64, -0.8 ,
                      -0.61, -0.88])
        )
    forecaster = ForecasterDirect(
                     estimator     = LinearRegression(),
                     steps         = 3,
                     lags          = 3,
                     transformer_y = StandardScaler(),
                     binner_kwargs = {'n_bins': 4}
                 )
    forecaster.fit(y=y, store_in_sample_residuals=True)
    results = forecaster._predict_interval_conformal(
        nominal_coverage=0.95, use_in_sample_residuals=True, use_binned_residuals=False
    )

    expected = pd.DataFrame(
                    data = np.array([
                            [-0.07720596, -2.17165565,  2.01724372],
                            [-0.54638907, -2.64083876,  1.54806061],
                            [-0.08892596, -2.18337565,  2.00552372]]),
                    index = pd.RangeIndex(start=20, stop=23, step=1),
                    columns = ['pred', 'lower_bound', 'upper_bound']
                )
    
    pd.testing.assert_frame_equal(results, expected)


def test_predict_interval_conformal_output_when_with_transform_y_and_transform_exog():
    """
    Test predict output when using LinearRegression as estimator, StandardScaler
    as transformer_y and transformer_exog as transformer_exog.
    """
    y = pd.Series(
            np.array([-0.59,  0.02, -0.9 ,  1.09, -3.61,  0.72, -0.11, -0.4 ,  0.49,
                       0.67,  0.54, -0.17,  0.54,  1.49, -2.26, -0.41, -0.64, -0.8 ,
                      -0.61, -0.88])
        )
    exog = pd.DataFrame(
               {'col_1': [7.5, 24.4, 60.3, 57.3, 50.7, 41.4, 87.2, 47.4, 36.5,
                          7.5, 24.4, 60.3, 57.3, 50.7, 41.4, 87.2, 47.4, 36.5,
                          7.5, 24.4],
                'col_2': ['a', 'a', 'a', 'a', 'a', 'b', 'b', 'b', 'b', 'b',
                          'a', 'a', 'a', 'a', 'a', 'b', 'b', 'b', 'b', 'b']}
           )
    exog_predict = exog.copy()
    exog_predict.index = pd.RangeIndex(start=20, stop=40)

    transformer_exog = ColumnTransformer(
                           [('scale', StandardScaler(), ['col_1']),
                            ('onehot', OneHotEncoder(), ['col_2'])],
                           remainder = 'passthrough',
                           verbose_feature_names_out = False
                       )
    forecaster = ForecasterDirect(
                     estimator        = LinearRegression(),
                     steps            = 5,
                     lags             = 3,
                     transformer_y    = StandardScaler(),
                     transformer_exog = transformer_exog
                 )
    forecaster.fit(y=y, exog=exog, store_in_sample_residuals=True)
    results = forecaster._predict_interval_conformal(
        steps=None, nominal_coverage=0.95, exog=exog_predict,
        use_in_sample_residuals=True, use_binned_residuals=False
    )

    expected = pd.DataFrame(
                   data = np.array([
                            [ 1.33676517, -0.63374319,  3.30727352],
                            [-1.05138096, -3.02188932,  0.91912739],
                            [ 0.55115225, -1.4193561 ,  2.52166061],
                            [ 0.86985865, -1.10064971,  2.84036701],
                            [ 0.44787213, -1.52263622,  2.41838049]]),
                   index = pd.RangeIndex(start=20, stop=25, step=1),
                   columns = ['pred', 'lower_bound', 'upper_bound']
               )
    
    pd.testing.assert_frame_equal(results, expected)


def test_predict_interval_conformal_output_when_forecaster_is_LinearRegression_steps_is_5_in_sample_residuals_is_True_binned_residuals_is_True():
    """
    Test output when estimator is LinearRegression 5 step ahead is predicted
    using in sample residuals.
    """
    forecaster = ForecasterDirect(
        estimator=LinearRegression(), steps=5, lags=3
    )
    forecaster.fit(y=y, store_in_sample_residuals=True)
    results = forecaster._predict_interval_conformal(
        steps=5, nominal_coverage=0.95, use_in_sample_residuals=True, use_binned_residuals=True
    )

    expected = pd.DataFrame(
                    data    = np.array(
                                [[ 0.51883519,  0.08813429,  0.94953609],
                                 [ 0.4584716 ,  0.04326184,  0.87368136],
                                 [ 0.39962743, -0.04758052,  0.84683537],
                                 [ 0.40452904, -0.0426789 ,  0.85173698],
                                 [ 0.41534557, -0.03186238,  0.86255351]]
                            ),
                    columns = ['pred', 'lower_bound', 'upper_bound'],
                    index   = pd.RangeIndex(start=50, stop=55, step=1)
                )

    pd.testing.assert_frame_equal(results, expected)


def test_predict_interval_conformal_output_when_forecaster_is_LinearRegression_steps_is_5_in_sample_residuals_is_False_binned_residuals_is_True():
    """
    Test output when estimator is LinearRegression, steps=5, use_in_sample_residuals=False,
    binned_residuals=True.
    """
    forecaster = ForecasterDirect(
        estimator=LinearRegression(), steps=5, lags=3
    )
    forecaster.fit(y=y, store_in_sample_residuals=True)
    forecaster.out_sample_residuals_by_bin_ = forecaster.in_sample_residuals_by_bin_
    results = forecaster._predict_interval_conformal(
        steps=5, nominal_coverage=0.95, use_in_sample_residuals=False, use_binned_residuals=True
    )

    expected = pd.DataFrame(
                    data    = np.array(
                                [[ 0.51883519,  0.08813429,  0.94953609],
                                 [ 0.4584716 ,  0.04326184,  0.87368136],
                                 [ 0.39962743, -0.04758052,  0.84683537],
                                 [ 0.40452904, -0.0426789 ,  0.85173698],
                                 [ 0.41534557, -0.03186238,  0.86255351]]
                            ),
                    columns = ['pred', 'lower_bound', 'upper_bound'],
                    index   = pd.RangeIndex(start=50, stop=55, step=1)
                )

    pd.testing.assert_frame_equal(results, expected)


def test_predict_interval_conformal_output_with_differentiation():
    """
    Test predict output when using differentiation.
    """
    forecaster = ForecasterDirect(
        estimator=LinearRegression(), steps=5, lags=3, differentiation=1
    )
    forecaster.fit(y=y, store_in_sample_residuals=True)
    results = forecaster._predict_interval_conformal(
        steps=5, nominal_coverage=0.95, use_in_sample_residuals=True, use_binned_residuals=False
    )

    expected = pd.DataFrame(
                    data    = np.array(
                                [[ 0.65659084,  0.11145843,  1.20172325],
                                    [ 0.6496844 , -0.44058042,  1.73994922],
                                    [ 0.61632235, -1.01907488,  2.25171958],
                                    [ 0.59545137, -1.58507827,  2.77598101],
                                    [ 0.64917089, -2.07649116,  3.37483294]]
                            ),
                    columns = ['pred', 'lower_bound', 'upper_bound'],
                    index   = pd.RangeIndex(start=50, stop=55, step=1)
                )

    pd.testing.assert_frame_equal(results, expected)



# Unit test get_feature_importances ForecasterStats
# ==============================================================================
import re
import pytest
import numpy as np
import pandas as pd
from skforecast.stats import Sarimax, Arar
from aeon.forecasting.stats import ARIMA
from skforecast.recursive import ForecasterStats
from sklearn.exceptions import NotFittedError

# Fixtures
from .fixtures_forecaster_stats import y
from .fixtures_forecaster_stats import exog


def test_NotFittedError_is_raised_when_forecaster_is_not_fitted():
    """
    Test NotFittedError is raised when calling get_feature_importances() and 
    forecaster is not fitted.
    """
    forecaster = ForecasterStats(estimator=Sarimax(order=(1, 0, 0)))

    err_msg = re.escape(
        ("This forecaster is not fitted yet. Call `fit` with appropriate "
         "arguments before using `get_feature_importances()`.")
    )
    with pytest.raises(NotFittedError, match=err_msg):         
        forecaster.get_feature_importances()


def test_output_get_feature_importances_ForecasterStats_with_Sarimax_estimator():
    """
    Test output of get_feature_importances ForecasterStats using Sarimax as
    estimator.
    """
    forecaster = ForecasterStats(
                     estimator = Sarimax(order= (1, 1, 1), maxiter=1000, method='cg', disp=False)
                 )
    forecaster.fit(y=y, exog=exog)
    results = forecaster.get_feature_importances(sort_importance=False)

    expected = pd.DataFrame({
                   'feature': ['exog', 'ar.L1', 'ma.L1', 'sigma2'],
                   'importance': np.array([0.9690539855149568, 0.4666537980992382, 
                                           -0.5263430267037418, 0.7862622654382363])
               })

    pd.testing.assert_frame_equal(expected, results)


def test_output_get_feature_importances_ForecasterStats_with_Arar_estimator():
    """
    Test output of get_feature_importances ForecasterStats using Arar as
    estimator.
    """
    forecaster = ForecasterStats(estimator = Arar(max_ar_depth=26, max_lag=40)) 
    forecaster.fit(y=y)
    results = forecaster.get_feature_importances(sort_importance=False)

    expected = pd.DataFrame(
                    {
                        "feature": {0: "lag_1", 1: "lag_10", 2: "lag_11", 3: "lag_12"},
                        "importance": {
                            0: 0.37984509460831944,
                            1: -0.208652315125212,
                            2: 0.23459105611032688,
                            3: -0.40956956673453687,
                        },
                    }
                )
    pd.testing.assert_frame_equal(expected, results)


def test_output_get_feature_importances_ForecasterStats_with_ARIMA_estimator():
    """
    Test output of get_feature_importances ForecasterStats using ARIMA as
    estimator.
    """
    forecaster = ForecasterStats(estimator = ARIMA(p=4, d=1, q=1))
    forecaster.fit(y=y)
    results = forecaster.get_feature_importances(sort_importance=True)

    expected = pd.DataFrame(
                {'feature': {0: 'lag_1',
                            1: 'lag_3',
                            2: 'intercept',
                            3: 'lag_4',
                            4: 'lag_2',
                            5: 'ma'
                        },
                'importance': {0: 0.6648894126315568,
                                1: 0.1478647645869075,
                                2: 0.0,
                                3: -0.23108776151763405,
                                4: -0.2444320861918562,
                                5: -0.8696258072896246
                        }
                }
            )
    pd.testing.assert_frame_equal(expected, results)



# Unit test select_features
# ==============================================================================
import re
import pytest
from sklearn.feature_selection import RFE
from sklearn.linear_model import LinearRegression
from skforecast.recursive import ForecasterRecursive
from skforecast.direct import ForecasterDirect
from skforecast.feature_selection import select_features
from skforecast.preprocessing import RollingFeatures

# Fixtures
from .fixtures_feature_selection import y_feature_selection as y
from .fixtures_feature_selection import exog_feature_selection as exog


def test_TypeError_select_features_raise_when_forecaster_is_not_supported():
    """
    Test TypeError is raised in select_features when forecaster is not supported.
    """
    forecaster = object()
    selector = RFE(estimator=LinearRegression(), n_features_to_select=3)

    err_msg = re.escape(
        "`forecaster` must be one of the following classes: ['ForecasterRecursive', "
        "'ForecasterDirect']."
    )
    with pytest.raises(TypeError, match = err_msg):
        select_features(
            selector   = selector,
            forecaster = forecaster,
            y          = y,
            exog       = exog,
        )


@pytest.mark.parametrize("select_only", 
                         ['not_exog_or_autoreg', 1, False], 
                         ids=lambda so: f'select_only: {so}')
def test_ValueError_select_features_select_only_not_autoreg_exog_None(select_only):
    """
    Test ValueError is raised in select_features when `select_only` is not 'autoreg',
    'exog' or None.
    """
    forecaster = ForecasterRecursive(
                     estimator = LinearRegression(),
                     lags      = 5,
                 )
    selector = RFE(estimator=forecaster.estimator, n_features_to_select=3)

    err_msg = re.escape(
        "`select_only` must be one of the following values: 'autoreg', 'exog', None."
    )
    with pytest.raises(ValueError, match = err_msg):
        select_features(
            selector    = selector,
            forecaster  = forecaster,
            y           = y,
            exog        = exog,
            select_only = select_only,
        )


@pytest.mark.parametrize("subsample", 
                         [-1, -0.5, 0, 0., 1.1, 2], 
                         ids=lambda ss: f'subsample: {ss}')
def test_ValueError_select_features_subsample_not_greater_0_less_equal_1(subsample):
    """
    Test ValueError is raised in select_features when `subsample` is not in (0, 1].
    """
    forecaster = ForecasterRecursive(
                     estimator = LinearRegression(),
                     lags      = 5,
                 )
    selector = RFE(estimator=forecaster.estimator, n_features_to_select=3)

    err_msg = re.escape(
        "`subsample` must be a number greater than 0 and less than or equal to 1."
    )
    with pytest.raises(ValueError, match = err_msg):
        select_features(
            selector   = selector,
            forecaster = forecaster,
            y          = y,
            exog       = exog,
            subsample  = subsample,
        )


def test_select_features_when_selector_is_RFE_and_select_only_is_exog_estimator():
    """
    Test that select_features returns the expected values when selector is RFE
    and select_only is 'exog' and estimator is passed to the selector instead
    of forecaster.estimator.
    """
    forecaster = ForecasterRecursive(
                     estimator = LinearRegression(),
                     lags      = 5,
                 )
    selector = RFE(estimator=LinearRegression(), n_features_to_select=3)

    selected_lags, selected_window_features, selected_exog = select_features(
        selector    = selector,
        forecaster  = forecaster,
        y           = y,
        exog        = exog,
        select_only = 'exog',
        verbose     = True,
    )

    assert selected_lags == [1, 2, 3, 4, 5]
    assert selected_window_features == []
    assert selected_exog == ['exog_0', 'exog_1', 'exog_2']


def test_select_features_when_selector_is_RFE_select_only_is_exog_ForecasterRecursive_no_window_features():
    """
    Test that select_features returns the expected values when selector is RFE
    and select_only is 'exog'. Forecaster is ForecasterRecursive and no window
    features are included.
    """
    forecaster = ForecasterRecursive(
                     estimator = LinearRegression(),
                     lags      = 5,
                 )
    selector = RFE(estimator=forecaster.estimator, n_features_to_select=3)

    selected_lags, selected_window_features, selected_exog = select_features(
        selector    = selector,
        forecaster  = forecaster,
        y           = y,
        exog        = exog,
        select_only = 'exog',
        verbose     = True,
    )

    assert selected_lags == [1, 2, 3, 4, 5]
    assert selected_window_features == []
    assert selected_exog == ['exog_0', 'exog_1', 'exog_2']


def test_select_features_when_selector_is_RFE_select_only_is_exog_ForecasterRecursive_window_features():
    """
    Test that select_features returns the expected values when selector is RFE
    and select_only is 'exog'. Forecaster is ForecasterRecursive and window
    features are included.
    """
    roll_features = RollingFeatures(
                        stats=['mean', 'std'],
                        window_sizes=[3, 5],
                    )
    forecaster = ForecasterRecursive(
                     estimator      = LinearRegression(),
                     lags           = 5,
                    window_features = roll_features
                 )
    selector = RFE(estimator=forecaster.estimator, n_features_to_select=3)

    selected_lags, selected_window_features, selected_exog = select_features(
        selector    = selector,
        forecaster  = forecaster,
        y           = y,
        exog        = exog,
        select_only = 'exog',
        verbose     = True,
    )

    assert selected_lags == [1, 2, 3, 4, 5]
    assert selected_window_features == ['roll_mean_3', 'roll_std_5']
    assert selected_exog == ['exog_0', 'exog_1', 'exog_2']


def test_select_features_when_selector_is_RFE_select_only_is_autoreg_ForecasterRecursive_no_window_features():
    """
    Test that select_features returns the expected values when selector is RFE
    and select_only is 'autoreg'. Forecaster is ForecasterRecursive and no window
    features are included.
    """
    forecaster = ForecasterRecursive(
                     estimator = LinearRegression(),
                     lags      = 5,
                 )
    selector = RFE(estimator=forecaster.estimator, n_features_to_select=3)

    selected_lags, selected_window_features, selected_exog = select_features(
        selector    = selector,
        forecaster  = forecaster,
        y           = y,
        exog        = exog,
        select_only = 'autoreg',
        verbose     = False,
    )

    assert selected_lags == [2, 3, 4]
    assert selected_window_features == []
    assert selected_exog == ['exog_0', 'exog_1', 'exog_2', 'exog_3', 'exog_4']


def test_select_features_when_selector_is_RFE_select_only_is_autoreg_ForecasterRecursive_window_features():
    """
    Test that select_features returns the expected values when selector is RFE
    and select_only is 'autoreg'. Forecaster is ForecasterRecursive and window
    features are used.
    """
    roll_features = RollingFeatures(
                        stats=['mean', 'std'],
                        window_sizes=[3, 5],
                    )
    forecaster = ForecasterRecursive(
                     estimator      = LinearRegression(),
                     lags           = 5,
                     window_features = roll_features
                 )
    selector = RFE(estimator=forecaster.estimator, n_features_to_select=4)

    selected_lags, selected_window_features, selected_exog = select_features(
        selector    = selector,
        forecaster  = forecaster,
        y           = y,
        exog        = exog,
        select_only = 'autoreg',
        verbose     = False,
    )

    assert selected_lags == [2, 3, 4]
    assert selected_window_features == ['roll_mean_3']
    assert selected_exog == ['exog_0', 'exog_1', 'exog_2', 'exog_3', 'exog_4']


def test_select_features_when_selector_is_RFE_select_only_is_None_ForecasterRecursive_no_window_features():
    """
    Test that select_features returns the expected values when selector is RFE
    and select_only is None. Forecaster is ForecasterRecursive and no window
    features are included.
    """
    forecaster = ForecasterRecursive(
                     estimator = LinearRegression(),
                     lags      = 5
                 )
    selector = RFE(estimator=forecaster.estimator, n_features_to_select=5)

    warn_msg = re.escape(
        "No autoregressive features have been selected. Since a Forecaster "
        "cannot be created without them, be sure to include at least one "
        "using the `force_inclusion` parameter."
    )
    with pytest.warns(UserWarning, match = warn_msg):
        selected_lags, selected_window_features, selected_exog = select_features(
            selector    = selector,
            forecaster  = forecaster,
            y           = y,
            exog        = exog,
            select_only = None,
            verbose     = False,
        )

    assert selected_lags == []
    assert selected_window_features == []
    assert selected_exog == ['exog_0', 'exog_1', 'exog_2', 'exog_3', 'exog_4']


def test_select_features_when_selector_is_RFE_select_only_is_None_ForecasterRecursive_window_features():
    """
    Test that select_features returns the expected values when selector is RFE
    and select_only is None. Forecaster is ForecasterRecursive and window
    features are included.
    """
    roll_features = RollingFeatures(
                        stats=['mean', 'std'],
                        window_sizes=[3, 5],
                    )
    forecaster = ForecasterRecursive(
                     estimator = LinearRegression(),
                     lags      = 5,
                     window_features = roll_features
                 )
    selector = RFE(estimator=forecaster.estimator, n_features_to_select=5)

    warn_msg = re.escape(
        "No autoregressive features have been selected. Since a Forecaster "
        "cannot be created without them, be sure to include at least one "
        "using the `force_inclusion` parameter."
    )
    with pytest.warns(UserWarning, match = warn_msg):
        selected_lags, selected_window_features, selected_exog = select_features(
            selector    = selector,
            forecaster  = forecaster,
            y           = y,
            exog        = exog,
            select_only = None,
            verbose     = False,
        )

    assert selected_lags == []
    assert selected_window_features == []
    assert selected_exog == ['exog_0', 'exog_1', 'exog_2', 'exog_3', 'exog_4']


def test_select_features_when_selector_is_RFE_select_only_autoreg_and_force_inclusion_is_regex():
    """
    Test that select_features returns the expected values when selector is RFE
    select_only is "autoreg" and force_inclusion is regex "^lag_".
    """
    roll_features = RollingFeatures(
                        stats=['mean', 'std'],
                        window_sizes=[3, 5],
                    )
    forecaster = ForecasterRecursive(
                     estimator = LinearRegression(),
                     lags      = 5,
                     window_features = roll_features
                 )
    selector = RFE(estimator=forecaster.estimator, n_features_to_select=3)

    selected_lags, selected_window_features, selected_exog = select_features(
        selector        = selector,
        forecaster      = forecaster,
        y               = y,
        exog            = exog,
        select_only     = 'autoreg',
        force_inclusion = "^lag_",
        verbose         = False,
    )

    assert selected_lags == [1, 2, 3, 4, 5]
    assert selected_window_features == []
    assert selected_exog == ['exog_0', 'exog_1', 'exog_2', 'exog_3', 'exog_4']


def test_select_features_when_selector_is_RFE_and_force_inclusion_is_regex():
    """
    Test that select_features returns the expected values when selector is RFE
    select_only is None and force_inclusion is regex "^roll_mean".
    """
    roll_features = RollingFeatures(
                        stats=['mean', 'std'],
                        window_sizes=[3, 5],
                    )
    forecaster = ForecasterRecursive(
                        estimator = LinearRegression(),
                        lags      = 5,
                        window_features = roll_features
                    )
    selector = RFE(estimator=forecaster.estimator, n_features_to_select=3)

    selected_lags, selected_window_features, selected_exog = select_features(
        selector        = selector,
        forecaster      = forecaster,
        y               = y,
        exog            = exog,
        select_only     = None,
        force_inclusion = "^roll_mean",
        verbose         = True,
    )

    assert selected_lags == []
    assert selected_window_features == ['roll_mean_3']
    assert selected_exog == ['exog_0', 'exog_1', 'exog_2']


def test_select_features_when_selector_is_RFE_select_force_inclusion_is_list():
    """
    Test that select_features returns the expected values when selector is RFE
    select_only  is None and force_inclusion is list.
    """
    forecaster = ForecasterRecursive(
                     estimator = LinearRegression(),
                     lags      = 5,
                 )
    selector = RFE(estimator=forecaster.estimator, n_features_to_select=3)

    selected_lags, selected_window_features, selected_exog = select_features(
        selector        = selector,
        forecaster      = forecaster,
        y               = y,
        exog            = exog,
        select_only     = None,
        force_inclusion = ['lag_1'],
        verbose         = False,
    )

    assert selected_lags == [1]
    assert selected_window_features == []
    assert selected_exog == ['exog_0', 'exog_1', 'exog_2']


def test_select_features_when_selector_is_RFE_select_only_is_exog_ForecasterDirect_no_window_features():
    """
    Test that select_features returns the expected values when selector is RFE
    and select_only is 'exog'. Forecaster is ForecasterDirect and no window
    features are included.
    """
    forecaster = ForecasterDirect(
                     estimator = LinearRegression(),
                     lags      = 5,
                     steps     = 3
                 )
    selector = RFE(estimator=forecaster.estimator, n_features_to_select=3)

    selected_lags, selected_window_features, selected_exog = select_features(
        selector    = selector,
        forecaster  = forecaster,
        y           = y,
        exog        = exog,
        select_only = 'exog',
        verbose     = True,
    )

    assert selected_lags == [1, 2, 3, 4, 5]
    assert selected_window_features == []
    assert selected_exog == ['exog_1', 'exog_2', 'exog_4']


def test_select_features_when_selector_is_RFE_select_only_is_exog_ForecasterDirect_window_features():
    """
    Test that select_features returns the expected values when selector is RFE
    and select_only is 'exog'. Forecaster is ForecasterDirect and window
    features are included.
    """
    roll_features = RollingFeatures(
                        stats=['mean', 'std'],
                        window_sizes=[3, 5],
                    )
    forecaster = ForecasterDirect(
                     estimator       = LinearRegression(),
                     lags            = 5,
                     window_features = roll_features,
                     steps           = 3
                 )
    selector = RFE(estimator=forecaster.estimator, n_features_to_select=3)

    selected_lags, selected_window_features, selected_exog = select_features(
        selector    = selector,
        forecaster  = forecaster,
        y           = y,
        exog        = exog,
        select_only = 'exog',
        verbose     = True,
    )

    assert selected_lags == [1, 2, 3, 4, 5]
    assert selected_window_features == ['roll_mean_3', 'roll_std_5']
    assert selected_exog == ['exog_1', 'exog_2', 'exog_4']



# Unit test predict_bootstrapping ForecasterDirectMultiVariate
# ==============================================================================
import re
import pytest
import numpy as np
import pandas as pd
from sklearn.exceptions import NotFittedError
from sklearn.linear_model import LinearRegression
from sklearn.compose import ColumnTransformer
from sklearn.preprocessing import StandardScaler
from sklearn.preprocessing import OneHotEncoder
from lightgbm import LGBMRegressor

from skforecast.preprocessing import RollingFeatures
from skforecast.preprocessing import TimeSeriesDifferentiator
from skforecast.direct import ForecasterDirectMultiVariate

# Fixtures
from .fixtures_forecaster_direct_multivariate import series
from .fixtures_forecaster_direct_multivariate import exog
from .fixtures_forecaster_direct_multivariate import exog_predict
from .fixtures_forecaster_direct_multivariate import data  # to test results when using differentiation

transformer_exog = ColumnTransformer(
                       [('scale', StandardScaler(), ['exog_1']),
                        ('onehot', OneHotEncoder(), ['exog_2'])],
                       remainder = 'passthrough',
                       verbose_feature_names_out = False
                   )


def test_predict_NotFittedError_when_fitted_is_False():
    """
    Test NotFittedError is raised when fitted is False.
    """
    forecaster = ForecasterDirectMultiVariate(
        estimator=LinearRegression(), level='l1', lags=3, steps=3
    )

    err_msg = re.escape(
        "This Forecaster instance is not fitted yet. Call `fit` with "
        "appropriate arguments before using predict."
    )
    with pytest.raises(NotFittedError, match = err_msg):
        forecaster.predict_bootstrapping(steps=5)


@pytest.mark.parametrize("use_binned_residuals", [True, False], 
                         ids=lambda binned: f'use_binned_residuals: {binned}')
def test_predict_bootstrapping_ValueError_when_out_sample_residuals_is_None(use_binned_residuals):
    """
    Test ValueError is raised when use_in_sample_residuals=False and
    forecaster.out_sample_residuals_ is None.
    """
    forecaster = ForecasterDirectMultiVariate(
        estimator=LinearRegression(), level='l1', lags=3, steps=2
    )
    forecaster.fit(series=series, store_in_sample_residuals=True)

    if use_binned_residuals:
        literal = "out_sample_residuals_by_bin_"
    else:
        literal = "out_sample_residuals_"

    err_msg = re.escape(
        f"`forecaster.{literal}` is either None or empty. Use "
        f"`use_in_sample_residuals = True` or the `set_out_sample_residuals()` "
        f"method before predicting."
    )
    with pytest.raises(ValueError, match = err_msg):
        forecaster.predict_bootstrapping(
            steps=1, use_in_sample_residuals=False, use_binned_residuals=use_binned_residuals
        )


@pytest.mark.parametrize("steps", [2, [1, 2], None], 
                         ids=lambda steps: f'steps: {steps}')
def test_predict_bootstrapping_output_when_in_sample_residuals_exog_and_transformer(steps):
    """
    Test output of predict_bootstrapping when estimator is LinearRegression,
    2 steps are predicted, using in-sample residuals, exog is included and both
    inputs are transformed.
    """
    forecaster = ForecasterDirectMultiVariate(
                     estimator          = LinearRegression(),
                     steps              = 2,
                     level              = 'l1',
                     lags               = 3,
                     transformer_series = StandardScaler(),
                     transformer_exog   = transformer_exog
                 )
    forecaster.fit(series=series, exog=exog, store_in_sample_residuals=True)
    results = forecaster.predict_bootstrapping(
        steps=steps, exog=exog_predict, n_boot=4, 
        use_in_sample_residuals=True, use_binned_residuals=False
    )
    
    expected = pd.DataFrame(
                   data = np.array([[0.68727676, 0.8016235 , 0.4586416 , 0.69518909],
                                    [0.07063188, 0.60208793, 0.20809564, 0.3706782 ]]),
                   columns = [f"pred_boot_{i}" for i in range(4)],
                   index   = pd.RangeIndex(start=50, stop=52)
               )
    expected.insert(0, 'level', np.tile(['l1'], 2))
    
    pd.testing.assert_frame_equal(expected, results)


def test_predict_bootstrapping_output_when_forecaster_is_LinearRegression_steps_is_2_in_sample_residuals_False_exog_and_transformer():
    """
    Test output of predict_bootstrapping when estimator is LinearRegression,
    2 steps are predicted, using in-sample residuals, exog is included and both
    inputs are transformed.
    """
    forecaster = ForecasterDirectMultiVariate(
                     estimator          = LinearRegression(),
                     steps              = 2,
                     level              = 'l1',
                     lags               = 3,
                     transformer_series = StandardScaler(),
                     transformer_exog   = transformer_exog
                 )
    forecaster.fit(series=series, exog=exog, store_in_sample_residuals=True)
    forecaster.out_sample_residuals_ = forecaster.in_sample_residuals_
    forecaster.out_sample_residuals_by_bin_ = forecaster.in_sample_residuals_by_bin_
    results = forecaster.predict_bootstrapping(
        steps=2, exog=exog_predict, n_boot=4, 
        use_in_sample_residuals=False, use_binned_residuals=False
    )
    
    expected = pd.DataFrame(
                   data = np.array([[0.68727676, 0.8016235 , 0.4586416 , 0.69518909],
                                    [0.07063188, 0.60208793, 0.20809564, 0.3706782 ]]),
                   columns = [f"pred_boot_{i}" for i in range(4)],
                   index   = pd.RangeIndex(start=50, stop=52)
               )
    expected.insert(0, 'level', np.tile(['l1'], 2))
    
    pd.testing.assert_frame_equal(expected, results)


def test_predict_bootstrapping_output_when_forecaster_is_LinearRegression_steps_is_2_in_sample_residuals_fixed():
    """
    Test output of predict_bootstrapping when estimator is LinearRegression,
    2 steps are predicted, using in-sample residuals that are fixed.
    """
    forecaster = ForecasterDirectMultiVariate(
                     estimator          = LinearRegression(),
                     steps              = 2,
                     level              = 'l1',
                     lags               = 3,
                     transformer_series = None
                 )
    forecaster.fit(series=series, exog=exog['exog_1'], store_in_sample_residuals=True)
    forecaster.in_sample_residuals_ = {'l1': np.array([1, 1, 1, 1, 1, 1, 1, 5, 5, 5, 5, 5, 5, 5])}
    results = forecaster.predict_bootstrapping(
        steps=2, exog=exog_predict['exog_1'], n_boot=4, 
        use_in_sample_residuals=True, use_binned_residuals=False
    )
    
    expected = pd.DataFrame(
                   data = np.array([[1.57457831, 5.57457831, 5.57457831, 1.57457831],
                                    [5.3777698 , 1.3777698 , 1.3777698 , 1.3777698 ]]),
                   columns = [f"pred_boot_{i}" for i in range(4)],
                   index   = pd.RangeIndex(start=50, stop=52)
               )
    expected.insert(0, 'level', np.tile(['l1'], 2))
    
    pd.testing.assert_frame_equal(expected, results)


def test_predict_bootstrapping_output_when_with_exog_and_differentiation_is_1_steps_1():
    """
    Test predict_bootstrapping output when using LinearRegression as estimator 
    and differentiation=1 and steps=1.
    """

    arr = data.to_numpy(copy=True)
    series_2 = pd.DataFrame(
        {'l1': arr,
         'l2': arr * 1.6},
        index=data.index
    )

    # Data differentiated
    differentiator = TimeSeriesDifferentiator(order=1)
    series_diff = pd.DataFrame(
        {'l1': differentiator.fit_transform(arr),
         'l2': differentiator.fit_transform(arr * 1.6)},
        index=data.index
    ).dropna()

    # Simulated exogenous variable
    rng = np.random.default_rng(9876)
    exog = pd.Series(
        rng.normal(loc=0, scale=1, size=len(data)), index=data.index, name='exog'
    )
    exog_diff = exog.iloc[1:]
    end_train = '2003-03-01 23:59:00'

    forecaster_1 = ForecasterDirectMultiVariate(
        estimator=LinearRegression(), level='l1', steps=1, lags=15, transformer_series=None
    )
    forecaster_1.fit(
        series=series_diff.loc[:end_train], exog=exog_diff.loc[:end_train], store_in_sample_residuals=True
    )
    boot_predictions_diff = forecaster_1.predict_bootstrapping(
        exog=exog_diff.loc[end_train:], n_boot=10, use_in_sample_residuals=True, use_binned_residuals=False
    )

    # Revert the differentiation
    last_value_train = series_2[['l1']].loc[:end_train].iloc[[-1]]
    boot_predictions_1 = boot_predictions_diff.drop(columns='level').copy()
    last_value_train = pd.DataFrame(
        np.full(shape=(1, 10), fill_value=last_value_train.values[0]), 
        columns = boot_predictions_1.columns, 
        index   = pd.date_range(last_value_train.index[0], periods=1)
    )
    boot_predictions_1 = pd.concat([last_value_train, boot_predictions_1])    
    boot_predictions_1 = boot_predictions_1.sort_index()
    boot_predictions_1 = boot_predictions_1.cumsum(axis=0).iloc[1:,]
    boot_predictions_1 = boot_predictions_1.asfreq('MS')
    boot_predictions_1.insert(0, 'level', np.tile(['l1'], 1))

    forecaster_2 = ForecasterDirectMultiVariate(
        estimator=LinearRegression(), level='l1', steps=1, lags=15, transformer_series=None, differentiation=1
    )
    forecaster_2.fit(
        series=series_2.loc[:end_train], exog=exog.loc[:end_train], store_in_sample_residuals=True
    )
    boot_predictions_2 = forecaster_2.predict_bootstrapping(
        exog=exog_diff.loc[end_train:], n_boot=10, use_in_sample_residuals=True, use_binned_residuals=False
    )

    pd.testing.assert_frame_equal(boot_predictions_1, boot_predictions_2)


def test_predict_bootstrapping_output_when_with_exog_and_differentiation_is_1_steps_10():
    """
    Test predict_bootstrapping output when using LinearRegression as estimator 
    and differentiation=1 and steps=10.
    """

    arr = data.to_numpy(copy=True)
    series_2 = pd.DataFrame(
        {'l1': arr,
         'l2': arr * 1.6},
        index=data.index
    )

    # Data differentiated
    differentiator = TimeSeriesDifferentiator(order=1)
    series_diff = pd.DataFrame(
        {'l1': differentiator.fit_transform(arr),
         'l2': differentiator.fit_transform(arr * 1.6)},
        index=data.index
    ).dropna()

    # Simulated exogenous variable
    rng = np.random.default_rng(9876)
    exog = pd.Series(
        rng.normal(loc=0, scale=1, size=len(data)), index=data.index, name='exog'
    )
    exog_diff = exog.iloc[1:]
    end_train = '2003-03-01 23:59:00'

    forecaster_1 = ForecasterDirectMultiVariate(
        estimator=LinearRegression(), level='l1', steps=10, lags=15, transformer_series=None
    )
    forecaster_1.fit(
        series=series_diff.loc[:end_train], exog=exog_diff.loc[:end_train], store_in_sample_residuals=True
    )
    boot_predictions_diff = forecaster_1.predict_bootstrapping(
        exog=exog_diff.loc[end_train:], n_boot=10, use_in_sample_residuals=True, use_binned_residuals=False
    )

    # Revert the differentiation
    last_value_train = series_2[['l1']].loc[:end_train].iloc[[-1]]
    boot_predictions_1 = boot_predictions_diff.drop(columns='level').copy()
    last_value_train = pd.DataFrame(
        np.full(shape=(1, 10), fill_value=last_value_train.values[0]), 
        columns = boot_predictions_1.columns, 
        index   = pd.date_range(last_value_train.index[0], periods=1)
    )
    boot_predictions_1 = pd.concat([last_value_train, boot_predictions_1])    
    boot_predictions_1 = boot_predictions_1.sort_index()
    boot_predictions_1 = boot_predictions_1.cumsum(axis=0).iloc[1:,]
    boot_predictions_1 = boot_predictions_1.asfreq('MS')
    boot_predictions_1.insert(0, 'level', np.tile(['l1'], 10))

    forecaster_2 = ForecasterDirectMultiVariate(
        estimator=LinearRegression(), level='l1', steps=10, lags=15, transformer_series=None, differentiation=1
    )
    forecaster_2.fit(
        series=series_2.loc[:end_train], exog=exog.loc[:end_train], store_in_sample_residuals=True
    )
    boot_predictions_2 = forecaster_2.predict_bootstrapping(
        exog=exog_diff.loc[end_train:], n_boot=10, use_in_sample_residuals=True, use_binned_residuals=False
    )

    pd.testing.assert_frame_equal(boot_predictions_1, boot_predictions_2)


def test_predict_output_when_window_features_steps_1():
    """
    Test output of predict when estimator is LGBMRegressor and window features
    with steps=1.
    """

    rolling = RollingFeatures(stats=['mean', 'sum'], window_sizes=[3, 5])
    forecaster = ForecasterDirectMultiVariate(
        estimator=LGBMRegressor(verbose=-1, random_state=123), level='l1', 
        steps=1, lags=5, window_features=rolling
    )
    forecaster.fit(series=series, exog=exog['exog_1'], store_in_sample_residuals=True)
    predictions = forecaster.predict_bootstrapping(
        n_boot=10, exog=exog_predict['exog_1'], use_in_sample_residuals=True, use_binned_residuals=False
    )

    expected = pd.DataFrame(
                   data    = np.array(
                                 [[0.60290414, 0.4125675 , 0.49927603, 0.64983046, 0.49009973,
                                   0.48971982, 0.40030064, 0.3430941 , 0.82287993, 0.67578235]]
                             ),
                   columns = [f"pred_boot_{i}" for i in range(10)],
                   index   = pd.RangeIndex(start=50, stop=51)
               )
    expected.insert(0, 'level', np.tile(['l1'], 1))
    
    pd.testing.assert_frame_equal(predictions, expected)


def test_predict_output_when_window_features_steps_10():
    """
    Test output of predict when estimator is LGBMRegressor and window features
    with steps=10.
    """

    rolling = RollingFeatures(stats=['mean', 'sum'], window_sizes=[3, 5])
    forecaster = ForecasterDirectMultiVariate(
        estimator=LGBMRegressor(verbose=-1, random_state=123), level='l1', 
        steps=10, lags=5, window_features=rolling
    )
    forecaster.fit(series=series, exog=exog['exog_1'], store_in_sample_residuals=True)
    predictions = forecaster.predict_bootstrapping(
        n_boot=10, exog=exog_predict['exog_1'], use_in_sample_residuals=True, use_binned_residuals=False
    )

    expected = pd.DataFrame(
                   data    = np.array(
                                 [[0.34317802, 0.64776959, 0.44252774, 0.72244338, 0.17452841,
                                    0.42501556, 0.38875996, 0.43489265, 0.66887236, 0.50222644],
                                [0.34888897, 0.1813542 , 0.24348837, 0.71499064, 0.3770138 ,
                                    0.44426026, 0.42297623, 0.6403034 , 0.31816366, 0.25635781],
                                [0.37107296, 0.39877449, 0.43370117, 0.61102351, 0.84943179,
                                    0.83690427, 0.24017543, 0.72904971, 0.43536741, 0.35150872],
                                [0.58509345, 0.83877703, 0.7137504 , 0.57562876, 0.34400862,
                                    0.52405127, 0.30293129, 0.96777975, 0.72154961, 0.17572187],
                                [0.24338025, 0.5347562 , 0.61422834, 0.61575098, 0.66106327,
                                    0.86630916, 0.31941155, 0.50577174, 0.39804426, 0.41112617],
                                [0.0988748 , 0.31547649, 0.83027159, 0.42189523, 0.40316123,
                                    0.71344514, 0.227533  , 0.60577608, 0.41454097, 0.63051606],
                                [0.74188031, 0.60966097, 0.17545176, 0.41697845, 0.4308126 ,
                                    0.06283256, 0.92888472, 0.41697845, 0.48408656, 0.60389823],
                                [0.54283777, 0.32254116, 0.44564736, 0.48303426, 0.30078917,
                                    0.42410373, 0.07449023, 0.32756542, 0.54283777, 0.54283777],
                                [0.33621667, 0.07766116, 0.51371272, 0.51509444, 0.61903383,
                                    0.64692848, 0.37116152, 0.3145088 , 0.70908819, 0.42496032],
                                [0.91812915, 0.47396699, 0.63571333, 0.34410137, 0.65201567,
                                    0.18341508, 0.46122504, 0.43951603, 0.33700121, 0.51837492]]
                             ),
                   columns = [f"pred_boot_{i}" for i in range(10)],
                   index   = pd.RangeIndex(start=50, stop=60)
               )
    expected.insert(0, 'level', np.tile(['l1'], 10))
    
    pd.testing.assert_frame_equal(predictions, expected)



# Unit test crps_from_predictions
# ==============================================================================
import re
import pytest
import numpy as np
from skforecast.metrics import crps_from_quantiles


def test_crps_from_quantiles_raises_error_when_y_true_is_not_int_or_float():
    """
    This test verifies that the function `crps_from_quantiles` raises a
    `TypeError` when the input `y_true` is not an integer or float.
    """

    y_true = "no valid input"
    pred_quantiles = np.array([1, 2, 3])
    quantile_levels = np.array([0.1, 0.5, 0.9])

    err_msg = re.escape("`y_true` must be a float or integer.")
    with pytest.raises(TypeError, match = err_msg):
        crps_from_quantiles(y_true, pred_quantiles, quantile_levels)


def test_crps_from_quantiles_raises_error_when_pred_quantiles_is_not_1d_array():
    """
    This test verifies that the function `crps_from_quantiles` raises a
    `TypeError` when the input `pred_quantiles` is not a 1D array.
    """
    y_true = 1
    pred_quantiles = "non valid input"
    quantile_levels = np.array([0.1, 0.5, 0.9])

    err_msg = re.escape("`pred_quantiles` must be a 1D numpy array.")
    with pytest.raises(TypeError, match = err_msg):
        crps_from_quantiles(y_true, pred_quantiles, quantile_levels)


def test_crps_from_quantiles_raises_error_when_quantile_levels_is_not_1d_array():
    """
    This test verifies that the function `crps_from_quantiles` raises a
    `TypeError` when the input `quantile_levels` is not a 1D array.
    """
    y_true = 1
    pred_quantiles = np.array([1, 2, 3])
    quantile_levels = "no valid input"

    err_msg = re.escape("`quantile_levels` must be a 1D numpy array.")
    with pytest.raises(TypeError, match = err_msg):
        crps_from_quantiles(y_true, pred_quantiles, quantile_levels)


def test_crps_from_quantiles_raises_error_when_quantile_levels_and_pred_quantiles_not_equal():
    """
    This test verifies that the function `crps_from_quantiles` raises a
    `ValueError` when the number of `quantile_levels` is not equal to the number
    of `pred_quantiles`.
    """
    y_true = 1
    pred_quantiles = np.array([1, 2, 3])
    quantile_levels = np.array([1, 2, 3, 4])

    err_msg = re.escape("The number of predicted quantiles and quantile levels must be equal.")
    with pytest.raises(ValueError, match = err_msg):
        crps_from_quantiles(y_true, pred_quantiles, quantile_levels)


def test_crps_from_quantiles_output():
    """
    This test verifies that the function `crps_from_quantiles` output.
    """
    y_true = 3.0
    quantile_levels = np.array([
        0.00, 0.025, 0.05, 0.10, 0.15, 0.20, 0.25, 0.30, 0.35, 0.40, 0.45, 0.50, 0.55,
        0.60, 0.65, 0.70, 0.75, 0.80, 0.85, 0.90, 0.95, 0.975, 1.00
    ])
    pred_quantiles = np.array([
        0.1, 1.0, 1.5, 2.0, 2.5, 3.0, 3.5, 4.0, 4.5, 5.0, 5.5, 6.0, 6.5, 7.0, 7.5,
        8.0, 8.5, 9.0, 9.5, 10.0, 10.5, 11.0, 11.5
    ])
    expected = 1.7339183102042313
    result = crps_from_quantiles(y_true, pred_quantiles, quantile_levels)

    np.testing.assert_almost_equal(result, expected)


# Unit test predict_interval ForecasterRecursive
# ==============================================================================
import re
import pytest
import numpy as np
import pandas as pd
from skforecast.recursive import ForecasterRecursive
from sklearn.compose import ColumnTransformer
from sklearn.preprocessing import StandardScaler
from sklearn.preprocessing import OneHotEncoder
from sklearn.linear_model import LinearRegression

# Fixtures
from .fixtures_forecaster_recursive import y


def test_check_interval_ValueError_when_method_is_not_valid_method():
    """
    Check ValueError is raised when `method` is not 'bootstrapping' or 'conformal'.
    """
    forecaster = ForecasterRecursive(LinearRegression(), lags=3)
    forecaster.fit(y=pd.Series(np.arange(10)))

    method = 'not_valid_method'
    err_msg = re.escape(
        f"Invalid `method` '{method}'. Choose 'bootstrapping' or 'conformal'."
    )
    with pytest.raises(ValueError, match = err_msg):
        forecaster.predict_interval(steps=1, method=method)


def test_predict_interval_output_when_forecaster_is_LinearRegression_steps_is_1_in_sample_residuals_is_True():
    """
    Test output when estimator is LinearRegression and one step ahead is predicted
    using in sample residuals.
    """
    forecaster = ForecasterRecursive(LinearRegression(), lags=3)
    forecaster.fit(y=pd.Series(np.arange(10)), store_in_sample_residuals=True)
    forecaster.in_sample_residuals_ = np.full_like(forecaster.in_sample_residuals_, fill_value=10)
    results = forecaster.predict_interval(
        steps=1, use_in_sample_residuals=True, use_binned_residuals=False
    )

    expected = pd.DataFrame(
                   data    = np.array([[10., 20., 20.]]),
                   columns = ['pred', 'lower_bound', 'upper_bound'],
                   index   = pd.RangeIndex(start=10, stop=11, step=1)
               )

    pd.testing.assert_frame_equal(results, expected)

    
def test_predict_interval_output_when_forecaster_is_LinearRegression_steps_is_2_in_sample_residuals_is_True():
    """
    Test output when estimator is LinearRegression and two step ahead is predicted
    using in sample residuals.
    """
    forecaster = ForecasterRecursive(LinearRegression(), lags=3)
    forecaster.fit(y=pd.Series(np.arange(10)), store_in_sample_residuals=True)
    forecaster.in_sample_residuals_ = np.full_like(forecaster.in_sample_residuals_, fill_value=10)
    results = forecaster.predict_interval(
        steps=2, use_in_sample_residuals=True, use_binned_residuals=False
    )

    expected = pd.DataFrame(
                   data    = np.array([[10., 20., 20.],
                                       [11., 24.33333333, 24.33333333]]),
                   columns = ['pred', 'lower_bound', 'upper_bound'],
                   index   = pd.RangeIndex(start=10, stop=12, step=1)
               )

    pd.testing.assert_frame_equal(results, expected)
    
    
def test_predict_interval_output_when_forecaster_is_LinearRegression_steps_is_1_in_sample_residuals_is_False():
    """
    Test output when estimator is LinearRegression and one step ahead is predicted
    using out sample residuals.
    """
    forecaster = ForecasterRecursive(LinearRegression(), lags=3)
    forecaster.fit(y=pd.Series(np.arange(10)), store_in_sample_residuals=True)
    forecaster.out_sample_residuals_ = np.full_like(forecaster.in_sample_residuals_, fill_value=10)
    results = forecaster.predict_interval(
        steps=1, use_in_sample_residuals=False, use_binned_residuals=False
    )

    expected = pd.DataFrame(
                   data    = np.array([[10., 20., 20.]]),
                   columns = ['pred', 'lower_bound', 'upper_bound'],
                   index   = pd.RangeIndex(start=10, stop=11, step=1)
               )

    pd.testing.assert_frame_equal(results, expected)
    
    
def test_predict_interval_output_when_forecaster_is_LinearRegression_steps_is_2_in_sample_residuals_is_False():
    """
    Test output when estimator is LinearRegression and two step ahead is predicted
    using out sample residuals.
    """
    forecaster = ForecasterRecursive(LinearRegression(), lags=3)
    forecaster.fit(y=pd.Series(np.arange(10)), store_in_sample_residuals=True)
    forecaster.out_sample_residuals_ = np.full_like(forecaster.in_sample_residuals_, fill_value=10)
    results = forecaster.predict_interval(
        steps=2, use_in_sample_residuals=False, use_binned_residuals=False
    )
    
    expected = pd.DataFrame(
                   data    = np.array([[10., 20., 20.],
                                       [11., 24.33333333, 24.33333333]]),
                   columns = ['pred', 'lower_bound', 'upper_bound'],
                   index   = pd.RangeIndex(start=10, stop=12, step=1)
               )

    pd.testing.assert_frame_equal(results, expected)


@pytest.mark.parametrize("interval", 
                         [0.90, (5, 95)], 
                         ids = lambda value: f'interval: {value}')
def test_predict_interval_output_when_with_transform_y(interval):
    """
    Test predict output when using LinearRegression as estimator and StandardScaler.
    """
    y = pd.Series(
            np.array([-0.59,  0.02, -0.9 ,  1.09, -3.61,  0.72, -0.11, -0.4 ,  0.49,
                       0.67,  0.54, -0.17,  0.54,  1.49, -2.26, -0.41, -0.64, -0.8 ,
                      -0.61, -0.88])
        )
    transformer_y = StandardScaler()
    forecaster = ForecasterRecursive(
                     estimator     = LinearRegression(),
                     lags          = 5,
                     transformer_y = transformer_y
                 )
    forecaster.fit(y=y, store_in_sample_residuals=False)
    forecaster.set_in_sample_residuals(y=y)
    results = forecaster.predict_interval(
        steps=5, interval=interval, use_in_sample_residuals=True, use_binned_residuals=False
    )

    expected = pd.DataFrame(
                   data = np.array([[-0.1578203 , -1.62232019,  1.55076389],
                                    [-0.18459942, -1.34560738,  0.70468515],
                                    [-0.13711051, -1.13938789,  1.49421349],
                                    [-0.01966358, -1.80104291,  1.07222337],
                                    [-0.03228613, -1.83428241,  1.15257045]]),
                   index = pd.RangeIndex(start=20, stop=25, step=1),
                   columns = ['pred', 'lower_bound', 'upper_bound']
               )
    
    pd.testing.assert_frame_equal(results, expected)


def test_predict_interval_output_when_with_transform_y_and_transform_exog():
    """
    Test predict output when using LinearRegression as estimator, StandardScaler
    as transformer_y and transformer_exog as transformer_exog.
    """
    y = pd.Series(
            np.array([-0.59, 0.02, -0.9, 1.09, -3.61, 0.72, -0.11, -0.4])
        )
    exog = pd.DataFrame(
               {'col_1': [7.5, 24.4, 60.3, 57.3, 50.7, 41.4, 87.2, 47.4],
                'col_2': ['a', 'a', 'a', 'a', 'b', 'b', 'b', 'b']}
           )
    exog_predict = exog.copy()
    exog_predict.index = pd.RangeIndex(start=8, stop=16)

    transformer_y = StandardScaler()
    transformer_exog = ColumnTransformer(
                            [('scale', StandardScaler(), ['col_1']),
                             ('onehot', OneHotEncoder(), ['col_2'])],
                            remainder = 'passthrough',
                            verbose_feature_names_out = False
                        )
    forecaster = ForecasterRecursive(
                     estimator        = LinearRegression(),
                     lags             = 5,
                     transformer_y    = transformer_y,
                     transformer_exog = transformer_exog
                 )
    
    forecaster.fit(y=y, exog=exog, store_in_sample_residuals=True)
    results = forecaster.predict_interval(
        steps=5, exog=exog_predict, interval=[5, 95], 
        use_in_sample_residuals=True, use_binned_residuals=False
    )

    expected = pd.DataFrame(
                   data = np.array([[ 0.50619336,  0.50619336,  0.50619336],
                                    [-0.09630298, -0.09630298, -0.09630298],
                                    [ 0.05254973,  0.05254973,  0.05254973],
                                    [ 0.12281153,  0.12281153,  0.12281153],
                                    [ 0.00221741,  0.00221741,  0.00221741]]),
                   index = pd.RangeIndex(start=8, stop=13, step=1),
                   columns = ['pred', 'lower_bound', 'upper_bound']
               )
    
    pd.testing.assert_frame_equal(results, expected)


def test_predict_interval_output_when_forecaster_is_LinearRegression_steps_is_5_in_sample_residuals_is_True_binned_residuals_is_True():
    """
    Test output when estimator is LinearRegression 5 step ahead is predicted
    using in sample binned residuals.
    """
    forecaster = ForecasterRecursive(LinearRegression(), lags=3, binner_kwargs={'n_bins': 15})
    forecaster.fit(y=y, store_in_sample_residuals=True)
    results = forecaster.predict_interval(
        steps=5, interval=(5, 95), use_in_sample_residuals=True, use_binned_residuals=True
    )

    expected = pd.DataFrame(
                    data    = np.array(
                                [[0.56842545, 0.25304752, 0.98664822],
                                [0.50873285, 0.05687183, 0.95641577],
                                [0.51189344, 0.23219067, 0.9780938 ],
                                [0.51559104, 0.16866501, 0.97647946],
                                [0.51060927, 0.12764843, 0.9697876 ]]
                            ),
                    columns = ['pred', 'lower_bound', 'upper_bound'],
                    index   = pd.RangeIndex(start=50, stop=55, step=1)
                )

    pd.testing.assert_frame_equal(results, expected)


def test_predict_interval_output_when_forecaster_is_LinearRegression_steps_is_5_in_sample_residuals_is_False_binned_residuals_is_True():
    """
    Test output when estimator is LinearRegression, steps=5, use_in_sample_residuals=False,
    binned_residuals=True.
    """
    forecaster = ForecasterRecursive(LinearRegression(), lags=3, binner_kwargs={'n_bins': 15})
    forecaster.fit(y=y, store_in_sample_residuals=True)
    forecaster.out_sample_residuals_by_bin_ = forecaster.in_sample_residuals_by_bin_
    results = forecaster.predict_interval(
        steps=5, interval=(5, 95), use_in_sample_residuals=False, use_binned_residuals=True
    )

    expected = pd.DataFrame(
                   data    = np.array(
                                [[0.56842545, 0.25304752, 0.98664822],
                                 [0.50873285, 0.05687183, 0.95641577],
                                 [0.51189344, 0.23219067, 0.9780938],
                                 [0.51559104, 0.16866501, 0.97647946],
                                 [0.51060927, 0.12764843, 0.9697876]]
                            ),
                   columns = ['pred', 'lower_bound', 'upper_bound'],
                   index   = pd.RangeIndex(start=50, stop=55, step=1)
               )
    
    pd.testing.assert_frame_equal(results, expected)


@pytest.mark.parametrize("interval", 
                         [0.95, (2.5, 97.5)], 
                         ids = lambda value: f'interval: {value}')
def test_predict_interval_conformal_output_when_with_transform_y(interval):
    """
    Test predict output when using LinearRegression as estimator and StandardScaler
    and conformal prediction.
    """
    y = pd.Series(
            np.array([-0.59,  0.02, -0.9 ,  1.09, -3.61,  0.72, -0.11, -0.4 ,  0.49,
                       0.67,  0.54, -0.17,  0.54,  1.49, -2.26, -0.41, -0.64, -0.8 ,
                      -0.61, -0.88])
        )
    forecaster = ForecasterRecursive(
                     estimator     = LinearRegression(),
                     lags          = 5,
                     transformer_y = StandardScaler()
                 )
    forecaster.fit(y=y, store_in_sample_residuals=True)
    results = forecaster.predict_interval(
        steps=5, method='conformal', interval=interval, 
        use_in_sample_residuals=True, use_binned_residuals=False
    )

    expected = pd.DataFrame(
                   data = np.array([
                              [-0.1578203 , -1.91279656,  1.59715596],
                              [-0.18459942, -1.93957568,  1.57037684],
                              [-0.13711051, -1.89208677,  1.61786574],
                              [-0.01966358, -1.77463983,  1.73531268],
                              [-0.03228613, -1.78726239,  1.72269012]]),
                   index = pd.RangeIndex(start=20, stop=25, step=1),
                   columns = ['pred', 'lower_bound', 'upper_bound']
               )
    
    pd.testing.assert_frame_equal(results, expected)


@pytest.mark.parametrize("interval", 
                         [0.95, (2.5, 97.5)], 
                         ids = lambda value: f'interval: {value}')
def test_predict_interval_conformal_output_when_binned_residuals(interval):
    """
    Test predict output when using LinearRegression as estimator and StandardScaler
    and conformal prediction with binned residuals.
    """
    y = pd.Series(
            np.array([-0.59,  0.02, -0.9 ,  1.09, -3.61,  0.72, -0.11, -0.4 ,  0.49,
                       0.67,  0.54, -0.17,  0.54,  1.49, -2.26, -0.41, -0.64, -0.8 ,
                      -0.61, -0.88])
        )
    forecaster = ForecasterRecursive(
                     estimator     = LinearRegression(),
                     lags          = 5,
                     transformer_y = StandardScaler(),
                     binner_kwargs = {'n_bins': 3}
                 )
    forecaster.fit(y=y, store_in_sample_residuals=True)
    results = forecaster.predict_interval(
        steps=5, method='conformal', interval=interval, 
        use_in_sample_residuals=True, use_binned_residuals=True
    )

    expected = pd.DataFrame(
                   data = np.array([
                              [-0.1578203 , -1.64020145,  1.32456084],
                              [-0.18459942, -1.66698056,  1.29778173],
                              [-0.13711051, -1.61949166,  1.34527063],
                              [-0.01966358, -1.50204472,  1.46271757],
                              [-0.03228613, -1.51466728,  1.45009501]]),
                   index = pd.RangeIndex(start=20, stop=25, step=1),
                   columns = ['pred', 'lower_bound', 'upper_bound']
               )
    
    pd.testing.assert_frame_equal(results, expected)



# Unit test fit ForecasterRecursiveClassifier
# ==============================================================================
import pytest
import numpy as np
import pandas as pd
from lightgbm import LGBMClassifier
from sklearn.linear_model import LogisticRegression
from skforecast.preprocessing import RollingFeaturesClassification
from skforecast.recursive import ForecasterRecursiveClassifier

# Fixtures
from .fixtures_forecaster_recursive_classifier import y, y_dt
from .fixtures_forecaster_recursive_classifier import exog, exog_dt


def custom_weights(index):  # pragma: no cover
    """
    Return 0 if index is between 20 and 40 else 1.
    """
    weights = np.where(
                (index >= 20) & (index <= 40),
                0,
                1
              )
    
    return weights


def test_forecaster_y_exog_features_stored():
    """
    Test forecaster stores y and exog features after fitting.
    """
    rolling = RollingFeaturesClassification(
        stats=['proportion', 'mode'], window_sizes=4
    )
    forecaster = ForecasterRecursiveClassifier(
        LogisticRegression(), lags=3, window_features=rolling
    )
    forecaster.fit(y=y, exog=exog)

    series_name_in_ = 'y'
    exog_in_ = True
    exog_type_in_ = type(exog)
    exog_names_in_ = ['exog']
    exog_dtypes_in_ = {'exog': exog.dtype}
    exog_dtypes_out_ = {'exog': exog.dtype}
    X_train_window_features_names_out_ = [
        'roll_proportion_4_class_0.0', 'roll_proportion_4_class_1.0',
        'roll_proportion_4_class_2.0', 'roll_mode_4'
    ]
    X_train_exog_names_out_ = ['exog']
    X_train_features_names_out_ = [
        'lag_1', 'lag_2', 'lag_3', 
        'roll_proportion_4_class_0.0', 'roll_proportion_4_class_1.0',
        'roll_proportion_4_class_2.0', 'roll_mode_4', 'exog'
    ]

    classes_ = [np.int64(1), np.int64(2), np.int64(3)]
    class_codes_ = [0.0, 1.0, 2.0]
    n_classes_ = 3
    encoding_mapping_ = {np.int64(1): 0.0, np.int64(2): 1.0, np.int64(3): 2.0}
    code_to_class_mapping_ = {0.0: np.int64(1), 1.0: np.int64(2), 2.0: np.int64(3)}
    
    assert forecaster.series_name_in_ == series_name_in_
    assert forecaster.exog_in_ == exog_in_
    assert forecaster.exog_type_in_ == exog_type_in_
    assert forecaster.exog_names_in_ == exog_names_in_
    assert forecaster.exog_dtypes_in_ == exog_dtypes_in_
    assert forecaster.exog_dtypes_out_ == exog_dtypes_out_
    assert forecaster.X_train_window_features_names_out_ == X_train_window_features_names_out_
    assert forecaster.X_train_exog_names_out_ == X_train_exog_names_out_
    assert forecaster.X_train_features_names_out_ == X_train_features_names_out_
    assert forecaster.classes_ == classes_
    assert forecaster.class_codes_ == class_codes_
    assert forecaster.n_classes_ == n_classes_
    assert forecaster.encoding_mapping_ == encoding_mapping_
    assert forecaster.code_to_class_mapping_ == code_to_class_mapping_


def test_forecaster_y_exog_features_stored_lgbm_as_categorical():
    """
    Test forecaster stores y and exog features after fitting when using 
    LGBMClassifier.
    """
    rolling = RollingFeaturesClassification(
        stats=['proportion', 'mode'], window_sizes=4
    )
    forecaster = ForecasterRecursiveClassifier(
        LGBMClassifier(verbose=-1), lags=3, window_features=rolling
    )
    forecaster.fit(y=y_dt, exog=exog_dt)

    series_name_in_ = 'y'
    exog_in_ = True
    exog_type_in_ = type(exog_dt)
    exog_names_in_ = ['exog']
    exog_dtypes_in_ = {'exog': exog_dt.dtype}
    exog_dtypes_out_ = {'exog': exog_dt.dtype}
    X_train_window_features_names_out_ = [
        'roll_proportion_4_class_0', 'roll_proportion_4_class_1',
        'roll_proportion_4_class_2', 'roll_mode_4'
    ]
    X_train_exog_names_out_ = ['exog']
    X_train_features_names_out_ = [
        'lag_1', 'lag_2', 'lag_3', 
        'roll_proportion_4_class_0', 'roll_proportion_4_class_1',
        'roll_proportion_4_class_2', 'roll_mode_4', 'exog'
    ]

    classes_ = [np.int64(1), np.int64(2), np.int64(3)]
    class_codes_ = [0, 1, 2]
    n_classes_ = 3
    encoding_mapping_ = {np.int64(1): 0, np.int64(2): 1, np.int64(3): 2}
    code_to_class_mapping_ = {0: np.int64(1), 1: np.int64(2), 2: np.int64(3)}
    
    assert forecaster.series_name_in_ == series_name_in_
    assert forecaster.exog_in_ == exog_in_
    assert forecaster.exog_type_in_ == exog_type_in_
    assert forecaster.exog_names_in_ == exog_names_in_
    assert forecaster.exog_dtypes_in_ == exog_dtypes_in_
    assert forecaster.exog_dtypes_out_ == exog_dtypes_out_
    assert forecaster.X_train_window_features_names_out_ == X_train_window_features_names_out_
    assert forecaster.X_train_exog_names_out_ == X_train_exog_names_out_
    assert forecaster.X_train_features_names_out_ == X_train_features_names_out_
    assert forecaster.classes_ == classes_
    assert forecaster.class_codes_ == class_codes_
    assert forecaster.n_classes_ == n_classes_
    assert forecaster.encoding_mapping_ == encoding_mapping_
    assert forecaster.code_to_class_mapping_ == code_to_class_mapping_


def test_forecaster_DatetimeIndex_index_freq_stored():
    """
    Test y_dt.index.freq is stored in forecaster.index_freq.
    """
    forecaster = ForecasterRecursiveClassifier(LogisticRegression(), lags=3)
    forecaster.fit(y=y_dt)
    expected = y_dt.index.freq
    results = forecaster.index_freq_

    assert results == expected


def test_forecaster_index_step_stored():
    """
    Test serie without DatetimeIndex, step is stored in forecaster.index_freq.
    """
    forecaster = ForecasterRecursiveClassifier(LogisticRegression(), lags=3)
    forecaster.fit(y=y)
    expected = y.index.step
    results = forecaster.index_freq_

    assert results == expected


@pytest.mark.parametrize("store_last_window", 
                         [True, False], 
                         ids=lambda lw: f'store_last_window: {lw}')
def test_fit_last_window_stored(store_last_window):
    """
    Test that values of last window are stored after fitting.
    """
    forecaster = ForecasterRecursiveClassifier(LogisticRegression(), lags=3)
    forecaster.fit(y=pd.Series(
        np.arange(50)), store_last_window=store_last_window
    )
    expected = pd.DataFrame(
        np.array([47, 48, 49]), index=[47, 48, 49], columns=['y']
    )

    if store_last_window:
        pd.testing.assert_frame_equal(forecaster.last_window_, expected)
    else:
        assert forecaster.last_window_ is None


def test_fit_model_coef_when_using_weight_func():
    """
    Check the value of the estimator coefs when using a `weight_func`.
    """
    forecaster = ForecasterRecursiveClassifier(
                     estimator   = LogisticRegression(),
                     lags        = 5,
                     weight_func = custom_weights
                 )
    forecaster.fit(y=y)
    results = forecaster.estimator.coef_
    expected = np.array(
        [
            [0.07295195, 0.26058532, -0.12877723, -0.70811789, 0.71544608],
            [0.16912783, -0.11071792, -0.00131558, 0.53663323, -0.83335583],
            [-0.24207979, -0.1498674, 0.13009281, 0.17148466, 0.11790975],
        ]
    )

    np.testing.assert_almost_equal(results, expected)


def test_fit_model_coef_when_not_using_weight_func():
    """
    Check the value of the estimator coefs when not using a `weight_func`.
    """
    forecaster = ForecasterRecursiveClassifier(
        estimator=LogisticRegression(), lags=5
    )
    forecaster.fit(y=y)
    results = forecaster.estimator.coef_
    expected = np.array(
        [
            [0.33080368, 0.01291936, -0.31347922, -0.43600223, 0.25260526],
            [-0.3486165, 0.05728672, -0.06652468, 0.25286034, -0.33176274],
            [0.01781282, -0.07020608, 0.38000389, 0.18314188, 0.07915748],
        ]
    )

    np.testing.assert_almost_equal(results, expected)



# Unit test check_one_step_ahead_input
# ==============================================================================
import re
import pytest
import numpy as np
import pandas as pd
from sklearn.linear_model import Ridge
from skforecast.exceptions import OneStepAheadValidationWarning
from skforecast.recursive import ForecasterRecursive
from skforecast.direct import ForecasterDirect
from skforecast.recursive import ForecasterRecursiveMultiSeries
from skforecast.direct import ForecasterDirectMultiVariate
from skforecast.model_selection._split import OneStepAheadFold
from skforecast.model_selection._utils import check_one_step_ahead_input

# Fixtures
from skforecast.model_selection.tests.fixtures_model_selection import y
from skforecast.model_selection.tests.fixtures_model_selection_multiseries import (
    series_dict_range,
    series_dict_dt
)


def test_check_one_step_ahead_input_TypeError_when_cv_not_OneStepAheadFold():
    """
    Test TypeError is raised in check_one_step_ahead_input if `cv` is not a
    OneStepAheadFold object.
    """
    forecaster = ForecasterRecursive(estimator=Ridge(), lags=2)
    y = pd.Series(np.arange(50))
    y.index = pd.date_range(start='2000-01-01', periods=len(y), freq='D')
    
    class BadCv():
        pass

    err_msg = re.escape("`cv` must be a 'OneStepAheadFold' object. Got 'BadCv'.")
    with pytest.raises(TypeError, match = err_msg):
        check_one_step_ahead_input(
            forecaster        = forecaster,
            cv                = BadCv(),
            metric            = 'mean_absolute_error',
            y                 = y,
            show_progress     = False,
            suppress_warnings = False
        )


def test_check_one_step_ahead_input_TypeError_when_forecaster_not_allowed():
    """
    Test TypeError is raised in check_one_step_ahead_input if `forecaster` is not
    allowed for OneStepAheadFold.
    """
    y = pd.Series(np.arange(50))
    y.index = pd.date_range(start='2000-01-01', periods=len(y), freq='D')

    forecasters_one_step_ahead = [
        "ForecasterRecursive",
        "ForecasterDirect",
        'ForecasterRecursiveClassifier',
        'ForecasterRecursiveMultiSeries',
        'ForecasterDirectMultiVariate'
    ]
    
    class BadForecaster():
        pass

    forecaster = BadForecaster()
    cv = OneStepAheadFold(initial_train_size=12)

    err_msg = re.escape(
        f"Only forecasters of type {forecasters_one_step_ahead} are allowed "
        f"when using `cv` of type `OneStepAheadFold`. Got {type(forecaster).__name__}."
    )
    with pytest.raises(TypeError, match = err_msg):
        check_one_step_ahead_input(
            forecaster        = forecaster,
            cv                = cv,
            metric            = 'mean_absolute_error',
            y                 = y,
            show_progress     = False,
            suppress_warnings = False
        )


@pytest.mark.parametrize("forecaster", 
                         [ForecasterRecursive(estimator=Ridge(), lags=2),
                          ForecasterDirect(estimator=Ridge(), lags=2, steps=3)], 
                         ids = lambda fr: f'{type(fr).__name__}')
def test_check_one_step_ahead_input_TypeError_when_y_is_not_pandas_Series(forecaster):
    """
    Test TypeError is raised in check_one_step_ahead_input if `y` is not a 
    pandas Series in forecasters uni-series.
    """
    bad_y = np.arange(50)
    cv = OneStepAheadFold(initial_train_size=len(bad_y) - 12)

    err_msg = re.escape(f"`y` must be a pandas Series. Got {type(bad_y)}")
    with pytest.raises(TypeError, match = err_msg):
        check_one_step_ahead_input(
            forecaster        = forecaster,
            cv                = cv,
            metric            = 'mean_absolute_error',
            y                 = bad_y,
            show_progress     = False,
            suppress_warnings = False
        )


@pytest.mark.parametrize("forecaster", 
                         [ForecasterDirectMultiVariate(estimator=Ridge(), lags=2, steps=3, level='l1')], 
                         ids = lambda fr: f'{type(fr).__name__}')
def test_check_one_step_ahead_input_TypeError_when_series_is_not_pandas_DataFrame(forecaster):
    """
    Test TypeError is raised in check_one_step_ahead_input if `series` is not a 
    pandas DataFrame in forecasters multiseries.
    """
    bad_series = pd.Series(np.arange(50))
    cv = OneStepAheadFold(initial_train_size=len(bad_series) - 12)

    err_msg = re.escape(f"`series` must be a pandas DataFrame. Got {type(bad_series)}")
    with pytest.raises(TypeError, match = err_msg):
        check_one_step_ahead_input(
            forecaster        = forecaster,
            cv                = cv,
            metric            = 'mean_absolute_error',
            series            = bad_series,
            show_progress     = False,
            suppress_warnings = False
        )


def test_check_one_step_ahead_input_TypeError_when_not_valid_exog_type():
    """
    Test TypeError is raised in check_one_step_ahead_input if `exog` is not a
    pandas Series, DataFrame or None.
    """
    y = pd.Series(np.arange(50))
    y.index = pd.date_range(start='2000-01-01', periods=len(y), freq='D')

    forecaster = ForecasterRecursive(estimator=Ridge(), lags=2)

    bad_exog = np.arange(50)
    
    cv = OneStepAheadFold(initial_train_size=len(y) - 12)

    err_msg = re.escape(
        f"`exog` must be a pandas Series, DataFrame or None. Got {type(bad_exog)}."
    )
    with pytest.raises(TypeError, match = err_msg):
        check_one_step_ahead_input(
            forecaster        = forecaster,
            cv                = cv,
            metric            = 'mean_absolute_error',
            y                 = y,
            exog              = bad_exog,
            show_progress     = False,
            suppress_warnings = False
        )


@pytest.mark.parametrize("differentiation", 
    [{'l1': 1, 'l2': 2, '_unknown_level': 1}, {'l1': 2, 'l2': None, '_unknown_level': 1}], 
     ids = lambda diff: f'differentiation: {diff}')
def test_check_one_step_ahead_input_ValueError_when_ForecasterRecursiveMultiSeries_diff_dict_not_cv_diff(differentiation):
    """
    Test ValueError is raised in check_one_step_ahead_input if `differentiation`
    of the ForecasterRecursiveMultiSeries as dict is different from 
    `differentiation` of the cv.
    """
    forecaster = ForecasterRecursiveMultiSeries(
        estimator=Ridge(), lags=2, differentiation=differentiation
    )
    cv = OneStepAheadFold(
        initial_train_size=len(series_dict_range['l1']) - 12, differentiation=1
    )
    
    err_msg = re.escape(
        "When using a dict as `differentiation` in ForecasterRecursiveMultiSeries, "
        "the `differentiation` included in the cv (1) must be "
        "the same as the maximum `differentiation` included in the forecaster "
        "(2). Set the same value "
        "for both using the `differentiation` argument."
    )
    with pytest.raises(ValueError, match = err_msg):
        check_one_step_ahead_input(
            forecaster        = forecaster,
            cv                = cv,
            metric            = 'mean_absolute_error',
            series            = series_dict_range,
            show_progress     = False,
            suppress_warnings = False
        )


@pytest.mark.parametrize("forecaster", 
    [ForecasterRecursive(estimator=Ridge(), lags=2, differentiation=2),
     ForecasterRecursiveMultiSeries(estimator=Ridge(), lags=2, differentiation=2)], 
     ids = lambda fr: f'{type(fr).__name__}')
def test_check_one_step_ahead_input_ValueError_when_forecaster_diff_not_cv_diff(forecaster):
    """
    Test ValueError is raised in check_one_step_ahead_input if `differentiation`
    of the forecaster is different from `differentiation` of the cv.
    """
    if type(forecaster).__name__ == 'ForecasterRecursive':
        data_length = len(y)
    else:
        data_length = len(series_dict_range['l1'])
    
    cv = OneStepAheadFold(initial_train_size=data_length - 12, differentiation=1)
    
    err_msg = re.escape(
        "The differentiation included in the forecaster "
        "(2) differs from the differentiation "
        "included in the cv (1). Set the same value "
        "for both using the `differentiation` argument."
    )
    with pytest.raises(ValueError, match = err_msg):
        check_one_step_ahead_input(
            forecaster        = forecaster,
            cv                = cv,
            metric            = 'mean_absolute_error',
            y                 = y,
            series            = series_dict_range,
            show_progress     = False,
            suppress_warnings = False
        )


def test_check_one_step_ahead_input_TypeError_when_metric_not_correct_type():
    """
    Test TypeError is raised in check_one_step_ahead_input if `metric` is not string, 
    a callable function, or a list containing multiple strings and/or callables.
    """
    forecaster = ForecasterRecursive(
                     estimator = Ridge(random_state=123),
                     lags      = 2
                 )
    
    cv = OneStepAheadFold(initial_train_size=len(y) - 12)
    
    metric = 5
    
    err_msg = re.escape(
        f"`metric` must be a string, a callable function, or a list containing "
        f"multiple strings and/or callables. Got {type(metric)}."
    )
    with pytest.raises(TypeError, match = err_msg):
        check_one_step_ahead_input(
            forecaster        = forecaster,
            cv                = cv,
            metric            = metric,
            y                 = y,
            show_progress     = False,
            suppress_warnings = False
        )


@pytest.mark.parametrize("initial_train_size", 
                         ['greater', 'smaller', 'date'], 
                         ids = lambda initial: f'initial_train_size: {initial}')
@pytest.mark.parametrize("forecaster", 
                         [ForecasterRecursive(estimator=Ridge(), lags=3),
                          ForecasterRecursiveMultiSeries(estimator=Ridge(), lags=3)], 
                         ids = lambda fr: f'{type(fr).__name__}')
def test_check_one_step_ahead_input_ValueError_when_initial_train_size_not_correct_value(initial_train_size, forecaster):
    """
    Test ValueError is raised in check_one_step_ahead_input when 
    initial_train_size >= length `y` or `series` or initial_train_size < window_size.
    """
    y_datetime = y.copy()
    y_datetime.index = pd.date_range(start='2020-01-01', periods=len(y), freq='D')

    if type(forecaster).__name__ == 'ForecasterRecursive':
        data_length = len(y_datetime)
        data_name = 'y'
    else:
        data_length = len(series_dict_dt['l1'])
        data_name = 'series'

    if initial_train_size == 'greater':
        initial_train_size = data_length
    elif initial_train_size == 'smaller':
        initial_train_size = forecaster.window_size - 1
    else:
        initial_train_size = '2020-01-02'  # Smaller than window_size
    
    cv = OneStepAheadFold(initial_train_size=initial_train_size)
    
    err_msg = re.escape(
        f"If `initial_train_size` is an integer, it must be greater than "
        f"the `window_size` of the forecaster ({forecaster.window_size}) "
        f"and smaller than the length of `{data_name}` ({data_length}). If "
        f"it is a date, it must be within this range of the index."
    )
    with pytest.raises(ValueError, match = err_msg):
        check_one_step_ahead_input(
            forecaster        = forecaster,
            cv                = cv,
            metric            = 'mean_absolute_error',
            y                 = y_datetime,
            series            = series_dict_dt,
            show_progress     = False,
            suppress_warnings = False
        )


@pytest.mark.parametrize("boolean_argument", 
                         ['show_progress', 'suppress_warnings'], 
                         ids = lambda argument: f'{argument}')
def test_check_one_step_ahead_input_TypeError_when_boolean_arguments_not_bool(boolean_argument):
    """
    Test TypeError is raised in check_one_step_ahead_input when boolean arguments 
    are not boolean.
    """
    forecaster = ForecasterRecursive(
                    estimator = Ridge(random_state=123),
                    lags      = 2
                 )
    
    cv = OneStepAheadFold(initial_train_size= len(y) - 12)
    
    boolean_arguments = {
        'show_progress': False,
        'suppress_warnings': False
    }
    boolean_arguments[boolean_argument] = 'not_bool'
    
    err_msg = re.escape(f"`{boolean_argument}` must be a boolean: `True`, `False`.")
    with pytest.raises(TypeError, match = err_msg):
        check_one_step_ahead_input(
            forecaster = forecaster,
            cv         = cv,
            metric     = 'mean_absolute_error',
            y          = y,
            **boolean_arguments
        )


def test_check_one_step_ahead_input_OneStepAheadValidationWarning():
    """
    Test OneStepAheadValidationWarning is warned in check_one_step_ahead_input
    when all checks are passed and suppress_warnings is False.
    """
    forecaster = ForecasterRecursive(
                     estimator = Ridge(random_state=123),
                     lags      = 2
                 )
    
    cv = OneStepAheadFold(initial_train_size=len(y) - 12)
    
    warn_msg = re.escape(
        "One-step-ahead predictions are used for faster model comparison, but they "
        "may not fully represent multi-step prediction performance. It is recommended "
        "to backtest the final model for a more accurate multi-step performance "
        "estimate."
    )
    with pytest.warns(OneStepAheadValidationWarning, match = warn_msg):
        check_one_step_ahead_input(
            forecaster        = forecaster,
            cv                = cv,
            metric            = 'mean_absolute_error',
            y                 = y,
            show_progress     = False,
            suppress_warnings = False
        )
